;; Generated package description from orgit.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "orgit" "2.1.2.0.20260301.125552" "Support for Org links to Magit buffers" '((emacs "28.1") (compat "30.1") (cond-let "0.2") (magit "4.5") (org "9.7")) :commit "4fb91faff3bf32dac5f6f932654c280cd1f190f7" :authors '(("Jonas Bernoulli" . "emacs.orgit@jonas.bernoulli.dev")) :maintainer '("Jonas Bernoulli" . "emacs.orgit@jonas.bernoulli.dev") :keywords '("hypermedia" "vc") :url "https://github.com/magit/orgit")
